/********************************************************************************
 * Copyright (c) 2023 CEA-List
 *
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License 2.0 which is available at
 * http://www.eclipse.org/legal/epl-2.0.
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 ********************************************************************************/

#ifndef AIDGE_CPU_OPERATOR_MULIMPL_KERNELS_H_
#define AIDGE_CPU_OPERATOR_MULIMPL_KERNELS_H_

#include "aidge/utils/Registrar.hpp"

#include <cstdint>     // std::int32_t, std::int64_t

#include "aidge/backend/cpu/data/Broadcasting.hpp"
#include "aidge/backend/cpu/operator/MulImpl.hpp"

namespace Aidge {
template <class I1, class I2, class O>
void MulImpl_cpu_forward_kernel(const std::vector<std::size_t>& input1Dims,
                                const std::vector<std::size_t>& input2Dims,
                                const std::vector<std::size_t>& outputDims,
                                const void* input1_,
                                const void* input2_,
                                void* output_) {

    const I1* input_1 = static_cast<const I1*>(input1_);
    const I2* input_2 = static_cast<const I2*>(input2_);
    O* output = static_cast<O*>(output_);

    size_t totalElements = 1;
    for (size_t dimSize : outputDims) {
        totalElements *= dimSize;
    }

	for (std::size_t oIndex = 0; oIndex < totalElements; ++oIndex)
	{
		std::vector<size_t> indexes = getMultiDimIndices(outputDims, oIndex);

		std::size_t idx1 = getFlattenedIndex(input1Dims, indexes);
		std::size_t idx2 = getFlattenedIndex(input2Dims, indexes);

        output[oIndex] = input_1[idx1] * input_2[idx2];
    }
}

template <class I1, class I2, class O>
void MulImpl_cpu_backward_kernel(const std::size_t input0Length, 
                                 const std::size_t input1Length,
                                 const std::size_t grad0Length,
                                 const std::vector<std::size_t> input0Dims,
                                 const std::vector<std::size_t> input1Dims,
                                 const void* input0_, 
                                 const void* input1_, 
                                 const void* grad_output_, 
                                 void* gradientInput0,
                                 void* gradientInput1)
{
    const auto* input0 = static_cast<const I1*>(input0_);
    const auto* input1 = static_cast<const I1*>(input1_);
    const auto* grad_output = static_cast<const O*>(grad_output_);
    auto* grad_input_0 = static_cast<I1*>(gradientInput0);
    auto* grad_input_1 = static_cast<I2*>(gradientInput1);


    if(input0Dims.size() >= input1Dims.size())
    {
        AIDGE_ASSERT(input0Length == grad0Length, "Incorrect dimensions between Mul input and output tensors");

        for(auto i = 0U; i < input0Length; ++i)
        {
            const auto indices = getMultiDimIndices(input1Dims, i);
            const auto flattenedIndex = getFlattenedIndex(input1Dims, indices);

            grad_input_0[i] = input1[flattenedIndex] * grad_output[i];
        }

        for(std::size_t i = 0 ; i < grad0Length; ++i)
        {
            const auto indices = getMultiDimIndices(input1Dims, i);
            const auto flattenedIndex = getFlattenedIndex(input1Dims, indices);

            grad_input_1[flattenedIndex] += input0[i] * grad_output[i];
        }

    } else {
        AIDGE_ASSERT(input1Length == grad0Length, "Incorrect dimensions between Mul input and output tensors");

        for(auto i = 0U; i < input1Length; ++i)
        {
            const auto indices = getMultiDimIndices(input0Dims, i);
            const auto flattenedIndex = getFlattenedIndex(input0Dims, indices);

            grad_input_1[i] = input0[flattenedIndex] * grad_output[i];
        }

        for(std::size_t i = 0 ; i < grad0Length; ++i)
        {
            const auto indices = getMultiDimIndices(input0Dims, i);
            const auto flattenedIndex = getFlattenedIndex(input0Dims, indices);

            grad_input_0[flattenedIndex] += input1[i] * grad_output[i];
        }
    }
}

// Kernels registration to implementation entry point
REGISTRAR(MulImpl_cpu,
    {DataType::Float32},
    {ProdConso::inPlaceModel, Aidge::MulImpl_cpu_forward_kernel<float, float, float>, Aidge::MulImpl_cpu_backward_kernel<float, float, float>});
REGISTRAR(MulImpl_cpu,
    {DataType::Float64},
    {ProdConso::inPlaceModel, Aidge::MulImpl_cpu_forward_kernel<double, double, double>, Aidge::MulImpl_cpu_backward_kernel<double, double, double>});
REGISTRAR(MulImpl_cpu,
    {DataType::Int32},
    {ProdConso::inPlaceModel, Aidge::MulImpl_cpu_forward_kernel<std::int32_t, std::int32_t, std::int32_t>, Aidge::MulImpl_cpu_backward_kernel<std::int32_t, std::int32_t, std::int32_t>});
REGISTRAR(MulImpl_cpu,
    {DataType::Int64},
    {ProdConso::inPlaceModel, Aidge::MulImpl_cpu_forward_kernel<std::int64_t, std::int64_t, std::int64_t>, Aidge::MulImpl_cpu_backward_kernel<std::int64_t, std::int64_t, std::int64_t>});
}  // namespace Aidge

#endif /* AIDGE_CPU_OPERATOR_MULIMPL_KERNELS_H_ */
